#!/usr/bin/env python3

from argparse import ArgumentParser, ArgumentTypeError, \
                     RawDescriptionHelpFormatter, BooleanOptionalAction
import sys
import os
import re
import pickle
from operator import add
from functools import reduce
import h5py
import numpy as np
import matplotlib.pyplot as plt

import gaussian_suite.gen_input as gsgi
import xyz_py as xyzp
from xyz_py import atomic
from molcas_suite.cli import ParseExtra
import gaussian_suite.cd_extractor as gscd
import hpc_suite as hpc
from hpc_suite.action import ParseKwargs, OrderedSelectionAction
from hpc_suite import make_parse_dict, SecondaryHelp
import angmom_suite
from angmom_suite import crystal
from angmom_suite.basis import Level
from angmom_suite.multi_electron import Ion
import molvis.core as mvis

from .coordinates import CoordinateInfo
from .derivative import print_tau_style, read_tau_style
from .lvc import LVC, generate_lvc_input, make_lvc_evaluator, ANG2BOHR, \
    LVCModelHamiltonian, LVCAdiabaticEnergies, LVCDiabaticHamiltonian
from .vibrations import Harmonic
from . import cells

plt.rcParams['font.family'] = "Arial"

HARTREE2INVCM = 219474.63 


# Action for secondary function help message
class FunctionHelp(SecondaryHelp):
    def __call__(self, parser, namespace, values, option_string=None):

        if namespace.function == 'cfp':
            angmom_suite.read_args(['cfp', '--help'])
        else:
            raise ValueError("Supply valid option to function argument.")


def str_int(x):
    try:
        return int(x)
    except ValueError:
        return str(x)


def coords_func(args):
    sp_data = CoordinateInfo.from_args(args)
    sp_data.to_file(args.distortion_info)

    return


def lvc_func(args):

    if args.add is not None:
        def join(a, b):
            return a.join(b, xyzp.load_xyz(args.reference_xyz)[1] * ANG2BOHR)
        lvc = reduce(join, (LVC.from_file(f_lvc) for f_lvc in args.add))

    elif args.efld is not None:
        charge_list = [np.loadtxt(f_chrg) for f_chrg in args.charge]
        xyz_list = [xyzp.load_xyz(f_xyz)[1] * ANG2BOHR for f_xyz in args.xyz]
        lvc = LVC.from_molcas_frozen_density(
            args.rassi, args.efld, charge_list, xyz_list)

    elif args.grad is not None:
        lvc = LVC.from_molcas(args.rassi, args.grad)

    else:
        raise ValueError("LVC gradient information supplied!")

    if args.active_atoms is not None:
        subset = [idx - 1 for idc in args.active_atoms for idx in idc]
        lvc = lvc.subset(subset)

    lvc.to_file(args.lvc_data, verbose=args.verbose)

    return


def eval_func(args, unknown_args):

    try:
        selected = args._selection
    except AttributeError:
        sys.exit("No quantity selected for evaluation!")

    # Resolve coordinates
    if args.vibration_info is not None:
        ho = Harmonic.from_file(args.vibration_info)
        coords = ho.mass_freq_weighted_coordinates * ANG2BOHR
    else:
        coords = None

    store_args = hpc.read_args(['store'] + unknown_args)

    eval_args = (store_args, make_lvc_evaluator, selected)
    eval_kwargs = {"order": args.order, "coords": coords,
                   "truncate": args.truncate, "align": args.align}

    if args.geom is None:
        hpc.store_func(*eval_args, geom=None, **eval_kwargs)
    else:
        for geom in map(lambda file: xyzp.load_xyz(file)[1], args.geom):
            hpc.store_func(*eval_args, geom=geom * ANG2BOHR, **eval_kwargs)


def vib_func(args):

    if args.gaussian_log:
        ho = Harmonic.from_gaussian_log(args.gaussian_log)
        if args.mass:
            raise ValueError("The --mass argument is not available with freq "
                             "data read from Gaussian text output use *.fchk!")

    elif args.gaussian_fchk:
        ho = Harmonic.from_gaussian_fchk(
            args.gaussian_fchk,
            ext_masses=args.mass,
            rot=args.rot,
            trans=args.trans
        )

    elif args.force_sets:
        ho = Harmonic.from_vasp_phonopy(
            args.poscar,
            args.force_sets,
            supercell=args.distortion_expansion,
            force_expansion=args.force_expansion,
            ext_masses=args.mass
        )

    subset = None if args.active_atoms is None else \
        [idx - 1 for idc in args.active_atoms for idx in idc]

    ho.subset_atoms(subset).to_file(args.vibration_info,
                                    store_vecs=args.store_vecs)

    if args.save_pyvibms is not None:
        ho.to_pyvibms(args.save_pyvibms)

    return


def prepare_func(args):

    if args.program == 'tau':
        # CFP args
        proj_kwargs = {
            'model_space': args.ground_level,
            'quax': args.quax,
            'terms': {"cf": [('J',)]},
            'k_max': args.k_max,
            'theta': True,
            'ion': args.ion
        }

        # EQ_CFPs.dat
        cfp_eq = LVCModelHamiltonian(LVC.from_file(args.lvc_data), 0,
                                     **proj_kwargs)[("cf", "J")]

        cfp_data = np.column_stack(
            (np.ones(27), list(cfp_eq.keys()), list(cfp_eq.values())))

        np.savetxt('EQ_CFPs.dat', cfp_data,
                   fmt=['%2d', '%2d', '% 2d', '%16.8f'])

        # mode_energies.dat
        ho = Harmonic.from_file(args.vibration_info)
        np.savetxt('mode_energies.dat', ho.freqs, fmt='%16.8f',
                   header="Frequency (cm-1)", comments='')

        cfp_dq = LVCModelHamiltonian(
            LVC.from_file(args.lvc_data), 1,
            coords=ho.mass_freq_weighted_coordinates * ANG2BOHR,
            **proj_kwargs
        )

        print_tau_style(filter(lambda pars: pars[0][1] == 'cf', iter(cfp_dq)),
                        ho.freqs, 'CFP_polynomials.dat')

    elif args.program == 'tau_direct':

        ho = Harmonic.from_file(args.vibration_info)
        np.savetxt('mode_energies.dat', ho.freqs, fmt='%16.8f',
                   header="Frequency (cm-1)", comments='')

        lvc = LVC.from_file(args.lvc_data)

        ener = LVCAdiabaticEnergies(lvc, 0, soc=True,
                                    truncate=args.truncate)[()]

        nstates = args.truncate or ener.size

        couplings = LVCDiabaticHamiltonian(lvc, 1, 
            coords=ho.mass_freq_weighted_coordinates * ANG2BOHR,
            soc=True, truncate=nstates)

        with h5py.File('tau.hdf5', 'w') as h:

            h.create_dataset('energies', data=(ener - ener[0]) * HARTREE2INVCM)

            grp = h.create_group("couplings")

            for modes, lin in iter(couplings):
                grp.create_dataset('_'.join(map(str, modes)),
                                   data=lin * HARTREE2INVCM)

            angm_grp = h.create_group('angmom')
            spin_grp = h.create_group('spin')

            for idx, comp in enumerate(["x", "y", "z"]):
                idc = np.ix_(range(nstates), range(nstates))
                angm_grp.create_dataset(comp, data=lvc.sos_angm[idx][idc])
                spin_grp.create_dataset(comp, data=lvc.sos_spin[idx][idc])
            
    return


def gen_charges_func(args):
    """
    Wrapper function for cli call to gen_charges
    """

    scell = cells.SuperCell.from_poscar(args.poscar, args.distortion_expansion,
                                        adjust_cutoff=args.adjust_cutoff)

    xyzp.save_xyz(
        "supercell.xyz", scell.atom_labels, scell.cart_coords, verbose=False
    )
    print()
    print("Supercell written to supercell.xyz")

    print("******************************")
    print("The unique entities are")
    for formula, indices in scell.asu_entities.items():
        print(formula)
    print("******************************")

    asu_indices = np.unique(scell.symmetry["equivalent_atoms"])

    xyzp.save_xyz(
        "unique_entities.xyz",
        scell.atom_labels[asu_indices],
        (scell.cart_coords + scell.repair_shifts)[asu_indices],
        verbose=False
    )
    print("Unique entities written to unique_entities.xyz\n")

    if len(asu_indices) < len(cells.flatten(scell.asu_entities.values())):
        print("\033[31m             ****Warning****") # noqa
        print("      High symmetry entity detected")
        print("Check calculated CHELPG charges obey symmetry \033[0m\n")

    # Create gaussian inputs for CHELPG calculation, one for each
    # member of the ASU

    # All electron aug-cc-pvtz
    ae_pvtz = [
        'H', 'Li', 'Be', 'Na', 'Mg', 'B', 'C', 'N', 'O', 'F', 'He', 'Ne', 'Al',
        'Si', 'P', 'S', 'Cl', 'Ar', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr'
    ]
    ae_basis = {
        "{}".format(atoms): "aug-cc-pvtz"
        for atoms in ae_pvtz
    }

    unsupported = [
        "K", "Ca", "Rb", "Sr", "Cs", "Ba", "Ac", "Th", "Pa", "U", "Np",
        "Pu", "Am", "Cm"
    ]
    unsupported_basis = {
        "{}".format(atoms): "????"
        for atoms in unsupported
    }

    for lab in scell.atom_labels[cells.flatten(scell.asu_entities.values())]:
        if lab in unsupported:
            print("\033[31m Warning, add basis for {} to .com file\033[0m \n".format(lab)) # noqa
    ae_basis = ae_basis | unsupported_basis

    # ECP
    pp_pvtz = [
        'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn'
    ]
    small_pp_basis = {
        "{}".format(atoms): "aug-cc-pvtz-pp"
        for atoms in pp_pvtz
    }

    stuttgart = {
        "{}".format(metals): "stuttgart rsc 1997"
        for metals in atomic.lanthanides + atomic.transition_metals
    }

    psuedo = small_pp_basis | stuttgart

    try:
        os.mkdir("gaussian")
    except OSError:
        pass

    for formula, indices in scell.asu_entities.items():
        gsgi.gen_input(
            "gaussian/{}.com".format(formula),
            scell.atom_labels[indices],
            (scell.cart_coords+scell.repair_shifts)[indices],
            99,
            99,
            method="PBE",
            bs_spec=ae_basis,
            ecp_spec=psuedo,
            opt=False,
            freq=False,
            extra_title="{}".format(formula),
            chelpg="charge"
        )
        print()

    print("\033[93mCharge and multiplicity set to 99 in all .com files")
    print("Please replace with actual values before submitting \033[0m \n")

    # Store supercell object in pickle for next step
    with open("supercell.pickle", 'wb') as _file:
        pickle.dump(scell, _file)

    print(
        "\033[0;32mSupercell object pickled in supercell.pickle \033[0m"
    )
    return


def parse_charges_func(args):
    """
    Wrapper function for cli call to parse_charges
    """

    # Load cell object from gen_charges
    with open(args.supercell_pickle, 'rb') as _file:
        scell: cells.Cell = pickle.load(_file)

    if args.calc_dir.lower() != 'cwd':
        calc_dir = "{}/".format(args.calc_dir)
    else:
        calc_dir = ""

    asu_charges = {
        head: gscd.get_chelpg_charges("{}/{}.log".format(calc_dir, head))
        for head in scell.asu_entities.keys()
    }

    index_to_charge = {
        index: charge
        for head in scell.asu_entities.keys()
        for index, charge in zip(scell.asu_entities[head], asu_charges[head])
    }

    index_to_basis_label = {
        index: "{}-{}".format(head, index + 1)
        for head in scell.asu_entities.keys()
        for index, charge in zip(scell.asu_entities[head], asu_charges[head])
    }

    supercell_charges = [
        index_to_charge[index] for index in scell.symmetry['equivalent_atoms']
    ]
    _min_charge = np.min(supercell_charges)
    _max_charge = np.max(supercell_charges)
    _charge_spread = _max_charge - _min_charge

    if args.vis_charges in ["entities", "all"]:

        viewers = []

        for it, indices in enumerate(scell.asu_entities.values()):
            # Get charge of current entity
            _charges = [supercell_charges[ind] for ind in indices]
            # Get labels and coords, and shift centre of mass to origin
            # so that interactive plot can be manipulated easily by user
            _labels = scell.atom_labels[indices]
            _coords = (scell.cart_coords+scell.repair_shifts)[indices]
            _coords -= xyzp.calculate_com(_labels, _coords)

            # Create Viewer object for current entity
            _tmp = cells.vis_charges_viewer(
                coords=_coords,
                labels=_labels,
                norm_charges=(_charges - _min_charge)/_charge_spread,
                viewer_style_args={
                    "height": "{}vh".format(95/len(asu_charges.keys())),
                    "width": "95vw"
                },
                viewer_div_args={
                    "div_id": "viewer_{:d}".format(it+1)
                }
            )
            viewers.append(_tmp)
        
        # Create webpage containing one viewer per entity with 
        # headers specifiying charge and colours
        body = [
            "<P> Maximum negative charge is {:f} e (Blue) </P>".format(
                _min_charge
            ),
            "<P> Maximum positive charge is {:f} e (Red) </P>".format(
                _max_charge
            ),
            "<P> Right click image to save </P>"
        ]
        body += [viewer.div for viewer in viewers]
        page = mvis.HtmlPage(
            scripts=[viewer.script for viewer in viewers],
            body=body
        )
        page.save("entities.html")
        print("\n\033[0;32mEntity charges can be visualised in entities.html\033[0m\n")

    central_atom = args.central_atom
    central_formula = args.central_formula

    # Pick user defined entity and move to centre of cell
    central_indices, without_central_indices = scell.find_centre_indices(
        central_formula=central_formula
    )

    catind = np.where(scell.atom_labels[central_indices] == central_atom)[0][0]
    catind = central_indices[catind]

    # Periodically recentre coordinates at central_formula
    scell.shift_centre_to(scell.frac_coords[catind])

    xyzp.save_xyz(
        "central_o_shifted.xyz",
        scell.atom_labels[central_indices],
        scell.cart_coords[central_indices]
    )

    xyzp.save_xyz(
        "environment_o_shifted.xyz",
        scell.atom_labels[without_central_indices],
        scell.cart_coords[without_central_indices]
    )

    # Make molcas basis file for environment charges
    _charges = [supercell_charges[wci] for wci in without_central_indices]
    cells.write_molcas_basis(
        [scell.atom_labels[index] for head in scell.asu_entities.keys()
                                  for index in scell.asu_entities[head]],
        {"{}-{}".format(ent, it + 1): chrg for ent in scell.asu_entities.keys()
         for it, chrg in zip(scell.asu_entities[ent], asu_charges[ent])},
        "ENV")

    supercell_labels = [
        "{}.ENV.{}".format(lab, index_to_basis_label[i]) if it not in central_indices else
        lab for it, (lab, i) in enumerate(zip(scell.atom_labels, scell.symmetry['equivalent_atoms']))]

    xyzp.save_xyz(
        "supercell_o_shifted.xyz", supercell_labels, scell.cart_coords)

    print("\033[93m")
    print("Molcas basis set file for environment written to:")
    print("     ENV (basis set specification)\033[0m")

    np.savetxt("environment_charges.dat", _charges)

    if args.vis_charges in ["supercell", "all"]:

        viewer = cells.vis_charges_viewer(
            coords=_coords,
            labels=_labels,
            norm_charges=(_charges - _min_charge)/_charge_spread,
            extra_coords=scell.cart_coords[central_indices],
            extra_labels=scell.atom_labels[central_indices],
            extra_color='#000000',
            viewer_style_args={
                "width":"95vw"
            },
            main_kwargs={"atom_opacities":0.5}
        )
        body = [
            "<P> Maximum negative charge is {:f} e (Blue) </P>".format(_min_charge),
            "<P> Maximum positive charge is {:f} e (Red) </P>".format(_max_charge),
            "<P> Right click image to save </P>",
            viewer.div
        ]
        page = mvis.HtmlPage(
            scripts=[viewer.script],
            body=body
        )

        page.save("supercell.html")
        print("\n\033[0;32mSupercell charges can be visualised in supercell.html\033[0m") #noqa

    return


def derivatives_func(args, unknown_args=None, **func_kwargs):

    if args.method == "findiff":
        dx = Finite.from_file(args.distortion_info)

    # parse extra function kwargs
    if not func_kwargs and unknown_args:
        if args.function == 'CFPs':
            func_kwargs = vars(angmom_suite.read_args(['cfp'] + unknown_args))
        elif not unknown_args:
            func_kwargs = {}
        else:
            raise ValueError("Unused extra arguments not known to the parser.")

    func = make_func(args, max_numerical_order=dx.order,
                     max_analytic_order=args.max_analytic_order, **func_kwargs)

    # prints derivatives and function values on the input grid
    if args.format == 'points':
        dx.write_points(
            args.function + "_points.dat", func, max_order=args.order)
        return

    ho = Harmonic.from_file(args.vibration_info)

    # remove normalisation and convert to units of zero point displacement
    conv = np.sqrt(ho.red_masses) * (ho.zpd if args.zpd_unit else 1.0)
    dQ = dx.transform(ho.displacements, func, args.order) * conv[:, np.newaxis]

    if args.format == 'tau':  # print central derivatives in tau style input

        if not args.order == 1:
            raise ValueError("Only first order derivatives with tau format.")

        if not args.function == 'CFPs':
            raise ValueError("Only function=CFPs possible with tau format.")

        print_tau_style(dQ, ho.freqs, 'CFP_polynomials.dat')

    else:  # print central derivatives to file
        pass

    return


def generate_lvc_input_func(args):

    # todo: make prettier by parsing numbers directly to list of ints using
    # parse_range and a custom action which flattens, check also other uses of
    # parse_range if flattened

    # todo: replace dict by global definition
    roots_dict = {'dy': [18]}

    if args.num_roots:
        num_roots = map(int, args.num_roots)
    elif args.ion:
        num_roots = roots_dict[args.ion.lower()]
    else:
        sys.exit("Invalid specification of the number of roots.")

    for iph_idx, num_root in enumerate(num_roots, start=1):

        if num_root == 0:
            continue

        generate_lvc_input(
            args.old_path,
            args.old_project,
            num_root,
            iph_idx,
            mclr_extra=args.mclr_extra,
            alaska_extra=args.alaska_extra,
            dry=args.dry
        )

    return


def parse_range(string):
    """
    adapted from https://stackoverflow.com/questions/6512280/accept-a-range-of-
    numbers-in-the-form-of-0-5-using-pythons-argparse
    """

    # match numbers in hyphen separated range
    # capture first and second string of digits, don't capture hyphon
    splt = re.match(r'(\d+)(?:-(\d+))?$', string)

    if not splt:
        raise ArgumentTypeError(
            "'{}' is not a range of number. Expected forms like '0-5' or '2'.".format( # noqa
                string
            )
        )

    # extract start and end of range, if single number is given -> end = start
    start = splt.group(1)
    end = splt.group(2) or start

    return list(range(int(start), int(end)+1))


def strength_func(args):
    """
    Wrapper for spin-phonon coupling strength CLI call
    """

    # Load CFP polynomials from file
    dQ, freqs = read_tau_style(args.cfp_file)

    # TODO delete
    args.n = 13
    args.J = 3.5
    args.L = 3
    args.S = 0.5

    if not args.nooef:
        # Get OEF values
        OEFs = crystal.calc_oef(args.n, args.J, args.L, args.S)
        print(np.shape(OEFs))

        # Add back in OEFs to polynomial values
        dQ *= OEFs

    # Calculate strength values of each mode
    S = np.array([crystal.calc_total_strength(mode) for mode in dQ])

    # Save strength to file
    np.savetxt("strengths.dat", S)
    print("Strengths saved to strengths.dat")

    # Read in symmetry labels if given
    if args.irreps:
        irreps = list(np.loadtxt(args.irreps, dtype=str))
    else:
        irreps = ['A']*len(freqs)

    unique_irreps, irrep_ints = np.unique(irreps, return_inverse=True)    

    if args.plot:

        _, ax = plt.subplots(num='Spin-phonon coupling strength')

        for unique_int in np.unique(irrep_ints):
            ax.stem(
                freqs[np.nonzero(irrep_ints == unique_int)[0]],
                S[np.nonzero(irrep_ints == unique_int)[0]],
                basefmt=' ',
                linefmt='C{:d}-'.format(unique_int),
                markerfmt='C{:d}o'.format(unique_int)
            )
        if args.irreps:
            ax.legend(unique_irreps, frameon=False)

        ax.set_ylabel(r'$S$ (cm$^{-1}$)', fontname="Arial")
        ax.set_xlabel('Mode energy (cm$^{-1}$)', fontname="Arial")

        ax.set_ylim([0., np.max(S)*1.05])

        plt.savefig("strengths.svg")
        plt.savefig("strengths.png")
        print("Strength plots saved to strengths.svg and strengths.png")
        plt.show()

    return


def read_args(arg_list=None):
    description = '''
    A package for performing Spin-Phonon coupling calculations.
    '''

    parser = ArgumentParser(
            description=description,
            formatter_class=RawDescriptionHelpFormatter
            )

    subparsers = parser.add_subparsers(dest='prog')

    coords = subparsers.add_parser('coords')
    coords.set_defaults(func=coords_func)

    coords.add_argument(
        '-D',
        '--distortion_info',
        type=str,
        help='HDF5 database containing information about the distortion.'
    )

    coords.add_argument(
        '--num_atoms',
        type=int,
        default=1,
        help='Number of atoms.'
    )

    coords.add_argument(
        '--mode_wise',
        nargs='*',
        default=None,
        type=parse_range,
        help='Mode indices to be included in distortions calculations ' +
             '- activates mode-wise distortions.'
        )

    coords.add_argument(
        '--atomic',
        nargs='*',
        default=None,
        type=parse_range,
        help='Atomic indices to be included in distortion calculations ' +
             '- activates atomic distortions.'
        )

    coords.add_argument(
        '--num_steps',
        type=int,
        default=0,
        help='Number of distortion steps.'
    )

    coords.add_argument(
        '--order',
        type=int,
        default=1,
        help='Order of distortion.'
    )

    coords.add_argument(
        '--constant_step',
        type=float,
        help='Step size factor for constant displacement.'
    )

    evaluate = subparsers.add_parser('eval')
    evaluate.set_defaults(func=eval_func)

    evaluate.add_argument(
        '-H',
        '--Help',
        action=FunctionHelp,
        help='show help message for additional arguments and exit'
    )

    evaluate.add_argument(
        '-L',
        '--lvc_data',
        type=str,
        help='HDF5 database containing the LVC parameters.'
    )

    coords = evaluate.add_mutually_exclusive_group()

    coords.add_argument(
        '-V',
        '--vibration_info',
        type=str,
        help=('HDF5 database containing information about the vibrations.'
              'Derivatives are evaluated in the basis of dimension-less'
              'mass-frequency weighted normal mode coordinates.')
    )

    evaluate.add_argument(
        '--geom',
        nargs='+',
        help='*.xyz coordinates at which properties will be evaluated.'
    )

    evaluate.add_argument(
        '--order',
        type=int,
        default=0,
        help='Order of derivative.'
    )

    evaluate.add_argument(
        '--truncate',
        type=int,
        metavar="max_state",
        help='Truncate matrix elements at max_state'
    )

    evaluate.add_argument(
        '--align',
        action=BooleanOptionalAction,
        default=True,
        help='Align LVC model reference to input geometry by minimising RMSD.'
    )

    evaluate.add_argument(
        '--adiabatic_energies',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['mch', 'soc'],
        help='Energies of the MCH or SOC eigenstates.'
    )

    evaluate.add_argument(
        '--diabatic_hamiltonian',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['mch', 'soc'],
        help='Diabatic potential energy matrix between MCH or SOC states.'
    )

    evaluate.add_argument(
        '--proj',
        nargs='+',
        action=OrderedSelectionAction,
        help='Model Hamiltonian projection'
    )

    evaluate.add_argument(
        '--sus',
        nargs='+',
        action=OrderedSelectionAction,
        help='Magnetic susceptibility'
    )

    inp = subparsers.add_parser('generate_input')
    inp.set_defaults(func=generate_lvc_input_func)

    inp.add_argument(
        'old_project',
        type=str,
        help='Project name of preceding Molcas calculation.'
    )

    inp.add_argument(
        '--old_path',
        type=str,
        default='../',
        help='Path to WorkDir of preceding Molcas calculation.'
    )

    roots = inp.add_mutually_exclusive_group(required=True)

    roots.add_argument(
        '--num_roots',
        nargs='+',
        help='Number of states per JOBIPH.'
    )

    roots.add_argument(
        '--ion',
        type=str,
        help='Label of the metal center, e.g. Dy.'
    )

    inp.add_argument(
        '--jobiph',
        nargs='+',
        help='Indices of Molcas JOBIPH wavefunction files *_IPH.'
    )

    inp.add_argument(
        '--mclr_extra',
        nargs='+',
        default=((), {}),
        type=make_parse_dict(str, str, key_only=True),
        action=ParseExtra,
        help='Manually run mclr with custom options, e.g. thre=1e-8',
        metavar='name=value')

    inp.add_argument(
        '--alaska_extra',
        nargs='+',
        default=((), {}),
        type=make_parse_dict(str, str, key_only=True),
        action=ParseExtra,
        help='Run alaska with custom options, e.g. cuto=1e-8',
        metavar='name=value')

    inp.add_argument(
        '--dry',
        default=False,
        action='store_true',
        help='Dry-run which prints files to be created'
    )

    lvc = subparsers.add_parser('lvc')
    lvc.set_defaults(func=lvc_func)

    lvc.add_argument(
        '-L', '--lvc_data',
        type=str,
        help='HDF5 database output containing the LVC data.'
    )

    grad = lvc.add_mutually_exclusive_group()

    grad.add_argument(
        '--grad',
        type=str,
        nargs='+',
        help='Molcas output file(s) containing gradients and NACs.'
    )

    grad.add_argument(
        '--efld',
        type=str,
        nargs='+',
        help='Molcas output file(s) containing electric field values.'
    )

    grad.add_argument(
        '--add',
        type=str,
        nargs='+',
        help='LVC HDF5 data bases to be added.'
    )

    frozen = lvc.add_argument_group('Frozen density gradient arguments')

    frozen.add_argument(
        '--charge',
        type=str,
        nargs='+',
        help='Text file(s) containing point charge values.'
    )

    frozen.add_argument(
        '--xyz',
        type=str,
        nargs='+',
        help='xyz file(s) containing point charge coordinates.'
    )

    add = lvc.add_argument_group('LVC addition arguments')

    add.add_argument(
        '--reference_xyz',
        type=str,
        help='xyz file containing the reference coordinates.'
    )

    lvc.add_argument(
        '--rassi',
        type=str,
        help=('Molcas *.rassi.h5 output file containing AMFI integrals, '
              'SF_ANGMOM operators and the spin multiplicities.')
    )

    lvc.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_range,
        help=(
            'Atomic indices active during spin-phonon coupling. Effectively '
            'effectively setting the coupling of all other atoms to zero. '
            'Useful to suppress coupling of specific atoms.'
        )
    )

    lvc.add_argument(
        '--verbose',
        action='store_true',
        help='Plot gradient norm and invariance measures.'
    )


    vibrate = subparsers.add_parser('vib')
    vibrate.set_defaults(func=vib_func)

    vibrate.add_argument(
        '-V', '--vibration_info',
        type=str,
        help='HDF5 database containing information about the vibrations.'
    )

    vib_calc_excl = vibrate.add_mutually_exclusive_group(required=True)

    vib_calc_excl.add_argument(
        '--gaussian_log',
        type=str,
        help='Text output of a gaussian freq calculation.'
    )

    vib_calc_excl.add_argument(
        '--gaussian_fchk',
        type=str,
        help='Formatted checkpoint file of a gaussian freq calculation.'
    )

    vib_calc_excl.add_argument(
        '--force_sets',
        type=str,
        help='FORCE_SETS file from VASP-phonopy pre-process.'
    )

    vibrate.add_argument(
        '--poscar',
        type=str,
        help='Unit cell POSCAR from VASP-phonopy pre-process.'
    )

    vibrate.add_argument(
        '--distortion_expansion',
        nargs=3,
        metavar=('N_x', 'N_y', 'N_z'),
        type=int,
        help='Supercell expansion.'
    )

    vibrate.add_argument(
        '--force_expansion',
        nargs=3,
        metavar=('N_x', 'N_y', 'N_z'),
        default=(1, 1, 1),
        type=int,
        help='Supercell expansion used in phonon calculation'
    )

    vibrate.add_argument(
        '--mass',
        type=make_parse_dict(str_int, float),
        default={},
        nargs='+',
        action=ParseKwargs,
        help='Modify atomic masses for isotopic substitution.',
        metavar='atom_index=mass or element_symbol=mass'
    )

    vibrate.add_argument(
        '--trans',
        action=BooleanOptionalAction,
        default=True,
        help='Project out three rigid body translations.'
    )

    vibrate.add_argument(
        '--rot',
        action=BooleanOptionalAction,
        default=True,
        help='Project out three rigid body rotations.'
    )

    vibrate.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_range,
        help=(
            'Atomic indices active during spin-phonon coupling. Effectively'
            ' subsets the displacement vectors. Useful if coupling is '
            'evaluated with a subset of the atoms present in the vibrational'
            ' calculation.'
        )
    )

    vibrate.add_argument(
        '--save_pyvibms',
        type=str,
        help='optional pyvibms output for visualisation in PyMol.'
    )

    vibrate.add_argument(
        '--store_vecs',
        action=BooleanOptionalAction,
        default=True,
        help='Flag to disable expensive storage of normal mode displacements.'
    )

    prepare = subparsers.add_parser('prep')
    prepare.set_defaults(func=prepare_func)

    prepare.add_argument(
        'program',
        choices=['tau', 'tau_direct'],
        help='Program for which to prepare inputs.'
    )

    data = prepare.add_argument_group('database files')

    data.add_argument(
        '-L', '--lvc_data',
        help='HDF5 database containing the LVC parameters.')

    data.add_argument(
        '-V', '--vibration_info',
        help='HDF5 database containing information about the vibrations.')

    prepare.add_argument(
        '--ground_level',
        type=Level.parse,
        help='Symbol of the model space.'
    )

    prepare.add_argument(
        '--ion',
        type=Ion.parse,
        help='Central ion.'
    )

    prepare.add_argument(
        '--k_max',
        type=int,
        default=6,
        help='Maximum Stevens operator rank.'
    )

    prepare.add_argument(
        '--quax',
        action=angmom_suite.QuaxAction,
        help='Quantisation axes.'
    )

    prepare.add_argument(
        '--truncate',
        type=int,
        metavar="max_state",
        help='Truncate matrix elements at max_state'
    )

    gen_charges = subparsers.add_parser(
        "gen_charges",
        description="""
        Creates inputs for Gaussian CHELPG charge calculations of each
        entity in the asymmetric unit of a VASP optimised structure
        """)
    gen_charges.set_defaults(func=gen_charges_func)

    gen_charges.add_argument(
        "poscar",
        type=str,
        help='Poscar containing optimised geometry'
    )

    gen_charges.add_argument(
        "distortion_expansion",
        type=int,
        nargs=3,
        help=(
            "Expansion used to generate supercell for spin phonon coupling"
            "calculations"
        )
    )

    gen_charges.add_argument(
        "--adjust_cutoff",
        type=make_parse_dict(str, float),
        default={},
        nargs='+',
        action=ParseKwargs,
        help=('Adjust cut-offs of certain elements in the construction of '
              'the adjacency matrix for the identification of unique entities '
              '(molecules) in the unit cell'),
        metavar='elem=cutoff'
    )

    parse_charges = subparsers.add_parser(
        "parse_charges",
        description=(
            "Collects result of Gaussian CHELPG charge calculations and maps"
            "charges back onto supercell coordinates."
        )
    )
    parse_charges.set_defaults(func=parse_charges_func)

    parse_charges.add_argument(
        "central_formula",
        type=str,
        help=(
            "Formula of central SMM or entity"
            "must match one of those given by gen_charges"
        )
    )

    parse_charges.add_argument(
        "central_atom",
        type=str,
        help='Central of central_formula, e.g. Dy'
    )

    parse_charges.add_argument(
        "--supercell_pickle",
        type=str,
        default='supercell.pickle',
        help='Supercell object pickle file'
    )

    parse_charges.add_argument(
        "--calc_dir",
        type=str,
        default='gaussian',
        help=(
            'Directory containing Gaussian .log file(s) with CHELPG charges'
            ' set to CWD for current directory'
        )
    )

    parse_charges.add_argument(
        "--vis_charges",
        type=str,
        help=(
            'Creates html file to visualise charges for entities and supercell'
        ),
        choices=("none", "all","supercell","entities")
    )

    strength = subparsers.add_parser('strength')
    strength.set_defaults(func=strength_func)

    strength.add_argument(
        "cfp_file",
        type=str,
        help=(
            "File (hdf5 or CFP_polynomials) containing coupling crystal",
            "field parameters"
        )
    )

    strength.add_argument(
        "n",
        type=float,
        help=(
            "Number of unpaired electrons in 4f subshell"
        )
    )

    strength.add_argument(
        "J",
        type=float,
        help=(
            "Total angular momentum quantum number"
        )
    )

    strength.add_argument(
        "L",
        type=float,
        help=(
            "Orbital angular momentum quantum number"
        )
    )

    strength.add_argument(
        "S",
        type=float,
        help=(
            "Spin angular momentum quantum number"
        )
    )

    strength.add_argument(
        "--plot",
        action="store_true",
        help="Produce plot of strength as a function of mode energy"
    )

    strength.add_argument(
        '--irreps',
        type=str,
        metavar='<file_name>',
        help=(
            'Color code strength plot based on mode symmetries listed in file',
            'file must contain column of IRREPs, one per mode'
        )
    )

    strength.add_argument(
        "--nooef",
        action="store_true",
        help="Produce plot of strength as a function of mode energy"
    )

    # If arg_list==None, i.e. normal cli usage, parse_args() reads from
    # "sys.argv". The arg_list can be used to call the argparser from the
    # back end.

    # read sub-parser
    parser.set_defaults(func=lambda args: parser.print_help())
    _args, _ = parser.parse_known_args(arg_list)

    # select parsing option based on sub-parser
    if _args.prog in ['derivatives', 'eval']:
        args, hpc_args = parser.parse_known_args(arg_list)
        args.func(args, hpc_args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)


def main():
    read_args()



from ase import neighborlist, Atoms
from ase.io import read
from ase.visualize import view
from phonopy.interface.vasp import read_vasp
from phonopy.structure.atoms import PhonopyAtoms
import numpy as np
import numpy.linalg as la
import xyz_py as xyzp
import spglib
import warnings
import itertools
import copy
import molvis.core as mvis
import matplotlib.cm as cm

def flatten(_list):
    """
    Flattens a list of lists by one subdimension

    e.g. [[1,2,3],[[2,3,4]]] ---> [1,2,3, [2,3,4]]

    Parameters
    ----------
    _list : list
        List of lists to be flattened

    Returns
    -------
    list
        Input list of lists flattened by one subdimension
    """

    _flist = list(itertools.chain(*_list))

    return _flist


class Molecule():

    """
    Contains structural information for a single molecule

    Attributes
    ----------

    formula : str
        Chemical formula as string
    indices : list
        Indices of each atom relative to supercell
    fragment_ind : list of lists
        List of lists, each containg a fragment of the molecule which is
        obtained when periodic boundary conditions are not considered
    dis_bonds : list of lists
        List of 2 element lists containing atomic indices specifying which
        bonds disappear when periodic boundary conditions are not considered

    """

    __slots__ = 'formula', 'indices', 'fragment_indices', 'dis_bonds'

    def __init__(self, formula, indices):

        self.formula = formula
        self.indices = indices
        self.fragment_indices = [[], [], []]
        self.dis_bonds = [[], [], []]

        return

    def add_fragments(self, axis, fragment_indices, dis_bonds):

        self.fragment_indices[axis] = fragment_indices
        self.dis_bonds[axis] = dis_bonds

        return


def calculate_molecule_shifts(molecule: Molecule, coords: list[list[float]],
                              lat_vecs: list[float]) -> list[float]:
    """
    Calculates translation operations required to reassemble
    current molecule from its constituent fragments

    Parameters
    ----------
    molecule : Molecule
        Molecule object for current molecule
    coords : list[list[float]]
        coordinates of full supercell
    lat_vecs : list[list[float]]
        lattice vectors of full supercell

    Returns
    -------
    list[float]
        Shift for each atom of cell in cartesian coordinates
    """

    shifts = np.zeros(np.shape(coords))

    n_atoms = len(molecule.indices)

    _atoms = Atoms(
        numbers=np.ones(n_atoms),
        positions=coords[molecule.indices],
        cell=lat_vecs,
        pbc=[True, True, True]
    )

    inds_to_order = {
        ind: it for it, ind in enumerate(molecule.indices)
    }

    pbc_distances = _atoms.get_all_distances(mic=True)

    for ax_it in range(3):
        for atom_a, atom_b in molecule.dis_bonds[ax_it]:

            pos_ps = np.where(shifts[atom_a] >= 1., 1., 0.)
            neg_ps = np.where(shifts[atom_a] <= -1., -1., 0.)

            pre_shift = (pos_ps + neg_ps) @ lat_vecs

            # Calculate distance between two atoms, accounting
            # for any shifts that are to be applied
            # This avoids double shifting back and forth
            dist = (coords + pre_shift)[atom_a, ax_it] - coords[atom_b, ax_it]

            # If distance is small now shifts have been applied then skip
            if np.abs(dist) < 1.05*pbc_distances[inds_to_order[atom_a], inds_to_order[atom_b]]: # noqa
                continue
            # Else, find other fragment indices and shift
            for fragi in molecule.fragment_indices[ax_it]:
                if atom_b in fragi:
                    fragi_b = fragi
            # Add 1 count of lattice vector to atom
            shifts[fragi_b, ax_it] += int(dist/np.abs(dist))

    # Avoid double shifting by the same lattice vector
    pos = np.where(shifts >= 1., 1., 0.)
    neg = np.where(shifts <= -1., -1., 0.)
    shifts = pos + neg

    cart_shifts = shifts @ lat_vecs

    return cart_shifts


class Cell():
    """
    Contains information on a given cell, unit or otherwise.

    Parameters
    ----------
    lat_vecs : np.array
        Lattice vectors as rows
    frac_coords : np.array
        Fractional coordinates of each atom in cell [a, b, c]
    atom_numbers : list[int]
        Atomic number of each atom in cell
    adjust_cutoff : dict
        Keys are element symbols values the new cut-offs

    Attributes
    ----------
    lat_vecs : np.array
        Lattice vectors as rows
    cell_params : list
        Cell parameters (norm of lattice vectors) [a, b, c]
    frac_coords : np.array
        Fractional coordinates of each atom in cell [a, b, c]
    cart_coords : np.array
        Cartesian coordinates of each atom
    atom_numbers : list[int]
        Atomic number of each atom in cell
    atom_labels : list[str]
        Atomic label of each atom in cell
    molecules : list[Molecules]
        cells.Molecule objects, one per complete entity in the unit cell
    symmetry : dict
        spglib.get_symmetry_dataset output
    repair_shifts : np.array
        Cartesian shifts which, when applied to the cartesian coordinates of
        the cell, repair the cell edges such that complete entities are present
        rather than fragments connected by periodic boundary conditions

    Returns
    -------
        None
    """

    __slots__ = [
        'lat_vecs', 'frac_coords', 'atom_numbers', 'cart_coords',
        'cell_params', 'atom_labels', 'atom_labels_nn', 'molecules',
        'symmetry', 'repair_shifts', 'asu_entities'
    ]

    def __init__(self, lat_vecs, frac_coords, atom_numbers,
                 adjust_cutoff=None):

        self.lat_vecs = lat_vecs
        self.frac_coords = frac_coords
        self.atom_numbers = atom_numbers

        self.cell_params = la.norm(self.lat_vecs, axis=1)

        # Convert atomic numbers to labels
        self.atom_labels = np.array(
            xyzp.num_to_lab(self.atom_numbers, numbered=False)
        )

        self.atom_labels_nn = np.array(
            xyzp.remove_label_indices(
                self.atom_labels
            )
        )

        # Convert fractional to cartesian coordinates
        self.cart_coords = self.frac_coords @ self.lat_vecs

        # Placeholder for list of molecules
        self.molecules = []

        # Find symmetry dataset
        self.symmetry = spglib.get_symmetry_dataset(
            (
                self.lat_vecs,
                self.frac_coords,
                self.atom_numbers
            )
        )

        # Find translations that repair edges of supercell to
        # give complete molecules
        self.repair_shifts = self.find_repair_shifts()

        # Find indices of whole, unique, entities in the asymmetric unit (ASU)
        # Entities are molecules, anions, cations
        # Keys of dictionary are formulae
        # e.g. H20_1, H20_2, ....
        # and values are list of indices of that entity relative to
        # supercell coordinates
        self.asu_entities = self.find_asu_entities(adjust_cutoff=adjust_cutoff)
        return

    @classmethod
    def from_phonopyatoms(cls, phonopy_atoms: PhonopyAtoms):
        """
        Creates cell object from PhonopyAtoms object
        """
        return cls(*phonopy_atoms.totuple())

    @classmethod
    def from_cif(cls, cif_file):
        """
        Creates cell object from cif file
        """

        # Ignore incorrect ase warning about space group
        warnings.filterwarnings("ignore", category=UserWarning)
        atoms = read(cif_file)

        return cls(
            atoms.cell,
            atoms.get_scaled_positions(),
            atoms.get_atomic_numbers()
        )

    def shift_centre_to(self,
                        new_frac_centre: list[float, float, float]) -> None:
        """
        Periodically shifts centre of cell to new point in fractional
        coordinates, and  recalculates cartesian coordinates to match

        Parameters
        ----------
        new_frac_centre : list[float, float, float]
            Fractional coordinates of new cell centre [a, b, c]

        Returns
        -------
            None
        """

        self.frac_coords += ([0.5, 0.5, 0.5] - new_frac_centre)
        self.frac_coords %= 1.
        self.cart_coords = self.frac_coords @ self.lat_vecs

        return

    @staticmethod
    def _get_adjacency(labels: list[str], coords: list, vecs: list,
                       pbc: list[bool, bool, bool] = [True, True, True],
                       visualise: bool = False) -> list:
        """
        Calculate adjacency matrix using ASE with or without periodic boundary
        conditions

        Parameters
        ----------
        labels : list
            Atomic labels
        coords : np.ndarray
            xyz coordinates as (n_atoms, 3) array
        vecs : np.ndarray
            Lattice vectors as (3,3) array
        pbc : list[bool], default [True, True, True]
            If true, apply periodic boundary conditions along specified cell
            axis [a, b, c]
        visualise : bool, default False
            If true, show cell in interactive 3d display

        Returns
        -------
        np.array
            Adjacency matrix with same order as labels/coords
        """

        # Make box
        box = Atoms(
            symbols=labels,
            positions=coords,
            cell=vecs,
            pbc=pbc
        )

        if visualise:
            view(box)

        # Define cutoffs for each atom using atomic radii
        cutoffs = neighborlist.natural_cutoffs(box)

        # Create neighbourlist using cutoffs
        neigh_list = neighborlist.NeighborList(
            cutoffs=cutoffs,
            self_interaction=False,
            bothways=True
        )

        # Update this list by specifying the atomic positions
        neigh_list.update(box)

        # Create adjacency matrix
        adjacency = neigh_list.get_connectivity_matrix(sparse=False)

        return adjacency

    def get_adjacency(self,
                      pbc: list[bool, bool, bool]
                      = [True, True, True]) -> list:
        """
        Calculate adjacency matrix using ASE with or without periodic boundary
        conditions

        Parameters
        ----------
        pbc : list[bool], default [True, True, True]
            If true, apply periodic boundary conditions along specified cell
            axis [a, b, c]
        Returns
        -------
        np.array
            Adjacency matrix with same order as labels/coords
        """

        adjacency = Cell._get_adjacency(
            self.atom_labels_nn,
            self.cart_coords,
            self.lat_vecs,
            pbc,
            visualise=False
        )

        return adjacency

    def get_entities(self,
                     pbc: list[bool, bool, bool] = [True, True, True]) -> dict:
        """
        Find entities in cell using ASE with or without periodic boundary
        conditions

        Parameters
        ----------
        pbc : list[bool], default [True, True, True]
            If true, apply periodic boundary conditions along specified cell
            axis [a, b, c]

        Returns
        -------
        dict
            keys = molecular formula,
            vals = list of lists, where each list contains the indices of a
                   single occurrence of the `key`, and the indices match the
                   order given in `labels` and `coords`

        """

        adjacency = self.get_adjacency(pbc)

        entities = xyzp.find_entities_from_adjacency(
            self.atom_labels_nn,
            adjacency
        )

        return entities

    def print_space_group(self):

        print("Structure is {}".format(self.symmetry["international"]))

        return

    def find_centre_indices(self, central_formula: str) -> tuple[list, list]:
        """
        Finds indices of entity with formula `central_formula`
        closest to centre of cell.

        Indices are given relative to full list of cell coordinates
        and atom labels

        Parameters
        ----------
        central_formula : str
            Formula string of central molecule

        Returns
        -------
        list
            Indices of central entity with `central_formula` in cell
        list
            Indices of environment (everything other than central entity)
            in cell
        """

        # Create adjacency matrix with periodic boundary conditions in all
        # directions
        entities_pbc = self.get_entities()

        # Indices of every instance of the central molecule
        all_cent_ent_indices = flatten(entities_pbc[central_formula])

        # Of the choices of central molecule, find the indices of the one
        # closest to the middle of the cell
        centre = np.sum(self.lat_vecs / 2., axis=1)
        closest_index = np.argmin(
            la.norm(
                self.cart_coords[all_cent_ent_indices] - centre,
                axis=1
            )
        )
        closest_index = all_cent_ent_indices[closest_index]

        environment_entities = copy.deepcopy(entities_pbc)

        for it, indices in enumerate(entities_pbc[central_formula]):
            if closest_index in indices:
                central_indices = indices
                environment_entities[central_formula].pop(it)

        environment_indices = flatten(
            flatten(
                environment_entities.values()
            )
        )

        return central_indices, environment_indices

    def find_repair_shifts(self):
        """
        Calculates translation operations required to "repair" molecules at
        the boundary of the supercell such that they are whole again
        """

        # Create adjacency matrix with periodic boundary conditions in all
        # directions
        adjacency_pbc = self.get_adjacency(
            pbc=[True, True, True]
        )

        # Find entities (molecules or otherwise) using adjacency matrix
        entities_pbc = xyzp.find_entities_from_adjacency(
            self.atom_labels_nn, adjacency_pbc
        )

        # Make list of molecules, which will tie fragments to full sets of
        # indices
        molecules = [
            Molecule(formula, full_mol)
            for formula, entities in entities_pbc.items()
            for full_mol in entities
        ]

        # Periodic boundary conditions deactivated for one of eac
        # direction
        pbc_choices = [
            [False, True, True],
            [True, False, True],
            [True, True, False],
        ]

        # Disable PBC in one cell direction, calculate adjacency, entities, and
        # find which bonds are cut
        for ax_it in range(3):
            adjacency_no_pbc = Cell._get_adjacency(
                self.atom_labels_nn, self.cart_coords, self.lat_vecs,
                pbc=pbc_choices[ax_it], visualise=False
            )

            entities_no_pbc = xyzp.find_entities_from_adjacency(
                self.atom_labels_nn, adjacency_no_pbc
            )

            # Find bonds which disappear when PBC disabled along current axis
            missing_bonds = np.array(
                np.nonzero(adjacency_pbc - adjacency_no_pbc)
            ).T

            entities_no_pbc_ind = flatten(entities_no_pbc.values())

            for molecule in molecules:

                # Collect fragments which make up current molecules
                fragments = [
                    [fra for fra in frag if fra in molecule.indices]
                    for frag in entities_no_pbc_ind
                ]
                # Remove None elements
                fragments = list(filter(None, fragments))
                # Find bonds which disappear when PBC deactivated
                dis_bonds = [
                    [atom_1, atom_2]
                    for atom_1, atom_2 in missing_bonds
                    if atom_1 in molecule.indices
                    and atom_2 in molecule.indices
                ]
                # Reorder fragment indices to have largest fragment first and
                # smallest last
                fragments = sorted(fragments, key=len, reverse=True)

                # Reorder disappearing bonds to be [large frag, small frag]
                tmp = []
                for atom_1, atom_2 in dis_bonds:
                    for frag in fragments:
                        if atom_1 in frag:
                            la1 = len(frag)
                        elif atom_2 in frag:
                            la2 = len(frag)
                    if la2 > la1:
                        tmp.append([atom_2, atom_1])
                    else:
                        tmp.append([atom_1, atom_2])
                dis_bonds = tmp

                molecule.add_fragments(ax_it, fragments, dis_bonds)

        self.molecules = molecules

        # For each molecule, check disconnected bonds and record shift
        # required to mend them. Shift is a lattice parameter
        repair_shift = np.zeros([self.atom_labels.size, 3])
        for molecule in molecules:
            if not len(molecule.dis_bonds):
                continue
            repair_shift += calculate_molecule_shifts(
                molecule, self.cart_coords, self.lat_vecs
            )

        return repair_shift

    def find_asu_entities(self, adjust_cutoff=None) -> dict:
        """
        Finds unique molecules of supercell by permuting unique atoms
        to give minimum number of molecules required to rebuild unit cell

        Returns
        -------
        dict
            keys are formula string of entity
            values are list of lists, each sublist containing indices of
            entity
        """

        asu_indices = np.unique(self.symmetry['equivalent_atoms'])

        n_unique = len(asu_indices)

        # Enforce that asymmetric unit atoms are of same molecules
        # since complete bonding has been traced, can just start with first
        # atom and go from there
        assignments = [False] * n_unique
        new_asu_indices = []

        asu_indices_to_assignment = {
            asu_index: it for it, asu_index in enumerate(asu_indices)
        }

        for it, asu_index in enumerate(asu_indices):
            # If this atom or one of its symmetrical equivalents
            # has already been included then skip
            if assignments[it]:
                continue

            # Else, find the molecule to which this unique atom belongs
            for molecule in self.molecules:
                if asu_index in molecule.indices:
                    # and add it to the new list of unique atoms
                    new_indices = copy.deepcopy(molecule.indices)

            # Update the assignment list to include new unique atoms
            for index in new_indices:
                new_asu_indices.append(index)
                assignments[
                    asu_indices_to_assignment[
                        self.symmetry["equivalent_atoms"][index]
                    ]
                ] = True

        asu_indices = new_asu_indices

        # Find entities in asymmetric unit
        # N.B. Indexing here is according to ASU subset, not full unit cell
        asu_ents = xyzp.find_entities(
            self.atom_labels_nn[asu_indices],
            (self.cart_coords + self.repair_shifts)[asu_indices],
            adjust_cutoff=adjust_cutoff
        )

        # Switch entity indexing to full unit cell
        new_asu_ents = {
            entity: [
                [asu_indices[ind] for ind in indices]
                for indices in full_indices
            ] for entity, full_indices in asu_ents.items()
        }

        # Create new list of equivalent atoms
        # using whole entities as unique atoms to which
        # all others are equivalent
        new_equiv = self.symmetry['equivalent_atoms']

        for new in asu_indices:
            old = self.symmetry['equivalent_atoms'][new]
            new_equiv[np.where(new_equiv == old)] = new

        self.symmetry['equivalent_atoms'] = new_equiv

        new_asu_ents = {
            "{}_{:d}".format(formula, it+1): indices
            for formula, all_indices in new_asu_ents.items()
            for it, indices in enumerate(all_indices)
        }

        return new_asu_ents

    def reconstruct_from_asu_unit(self):

        # Rotate lattice vectors into standard basis
        _standard_lat_vecs = (
            self.lat_vecs.T @ self.symmetry["transformation_matrix"]
        ).T

        # Shift origin of coordinate system
        _standard_frac_coords = np.array([
            coord + self.symmetry["origin_shift"]
            for coord in self.frac_coords
        ])

        # Rotate standard lattice vectors into idealised basis
        _ideal_lat_vecs = (
            self.symmetry["std_rotation_matrix"] @ _standard_lat_vecs.T
        ).T

        rotations = self.symmetry['rotations']
        translations = self.symmetry['translations']

        # Some sort of origin shift is required here but is not
        # performed by spglib
        # for it in range(306):
        #     self.frac_coords[it] += np.diag(
        # np.array([[-0.5, 0., 0.],[0, -0.5, 0.],[0., 0., -0.5]])
        # )

        asu_indices = np.unique(self.symmetry['equivalent_atoms'])

        recon_f_coords = np.vstack([
            (rot @ _standard_frac_coords[asu_indices].T).T + trans
            for rot, trans in zip(rotations, translations)
        ])

        recon_c_coords = recon_f_coords @ _ideal_lat_vecs

        recon_labels = [self.atom_labels_nn[asu_indices]]
        recon_labels *= rotations.shape[0]
        recon_labels = np.concatenate(recon_labels)

        return recon_c_coords, recon_labels


class SuperCell(Cell):
    """
    Contains information on supercell created from a given cell with
    specified expansion

    Parameters
    ----------
    lat_vecs : list
        Lattice vectors of unit cell as rows
    frac_coords : list
        Coordinates of each atom in cell as fraction of cell parameters
    atom_numbers : list
        Atomic number of each atom in cell
    expansion : list[int]
        Requested (integer) supercell expansion e.g. [3, 3, 3]

    Attributes
    ----------
    expansion : list[int]
        Expansion used to generate supercell from cell e.g. [3, 3, 3]
    uc_sc_map : list
        Mapping of supercell atoms onto unit cell atoms
    """
    def __init__(self, lat_vecs: list[float], frac_coords: list[float],
                 atom_numbers: list[int], expansion: list[float],
                 **kwargs) -> None:

        if any([expand < 0. for expand in expansion]):
            raise ValueError("Requested expansion is negative")

        self.expansion = expansion

        _frac_coords = copy.deepcopy(frac_coords)
        _atom_numbers = list(copy.deepcopy(atom_numbers))
        _lat_vecs = copy.deepcopy(lat_vecs)
        n_atoms = len(atom_numbers)

        unitvec = np.eye(3)

        # Expand along each cell axis
        for vec, expand in zip(unitvec, expansion):
            _frac_coords = [
                _frac_coords + vec*it
                for it in range(expand)
            ]
            _frac_coords = np.concatenate(_frac_coords)

        # Map from supercell to unit cell
        self.sc_uc_map = np.concatenate(
            [np.arange(n_atoms)]*np.prod(expansion)
        )

        # Atomic numbers
        _atom_numbers = np.concatenate(
            [atom_numbers]*np.prod(expansion)
        )

        # Convert fractional coords to cartesian using cell vectors
        _cart_coords = _frac_coords @ _lat_vecs

        # Make lattice vectors larger
        _lat_vecs = (_lat_vecs.T * expansion).T

        # Convert cartesian coordinates to fractional coordinates with new
        # larger lattice vectors
        # Convert fractional coords to cartesian using cell vectors
        _frac_coords = _cart_coords @ la.inv(_lat_vecs)

        return super().__init__(_lat_vecs, _frac_coords, _atom_numbers,
                                **kwargs)

    @classmethod
    def from_poscar(cls, poscar_name, expansion, adjust_cutoff=None):
        atoms = read_vasp(poscar_name)
        return cls(atoms.cell, atoms.scaled_positions, atoms.numbers,
                   expansion, adjust_cutoff=adjust_cutoff)


def write_molcas_basis(labels, charge_dict, name):
    """
    Writes dummy molcas basis file for environment charges to a textfile named
    according to the basis name.

    Parameters
    ----------
    labels : list[str]
        Atomic labels of environment with no indexing
    charge_dict : dict
        CHELPG charge of each environment atom
    name : str
        Root name of basis
    """

    labels = xyzp.remove_label_indices(labels)

    with open(name, 'w') as f:

        f.write("* This file was generated by spin_phonon_suite\n")
        for elem, (lab, chrg) in zip(labels, charge_dict.items()):
            f.write(f"/{elem}.{name}.{lab}.0s.0s.\n")
            f.write("Dummy basis set for atomic charges of environment\n")
            f.write("no ref\n")
            f.write(f"{chrg:.9f} 0\n")
            f.write("0 0\n")

    return


def vis_charges_viewer(coords, labels, norm_charges, extra_coords=[],
                       extra_labels=[], extra_color='', viewer_style_args={},
                       viewer_div_args={}, main_kwargs={}, extra_kwargs={}):

    ms1 = mvis.Model(labels, coords, **main_kwargs)

    colours = cm.get_cmap('coolwarm', 250)

    ms1.atom_colours = [
        '#{0:02x}{1:02x}{2:02x}'.format(*col[:-1])
        for col in colours(norm_charges, bytes=True)
    ]

    ms2 = mvis.Model(extra_labels, extra_coords, **extra_kwargs)

    if extra_color:
        ms2.atom_colours = extra_color

    viewer = mvis.Viewer(
        objects=[ms1, ms2],
        extra_div_args=viewer_div_args,
        extra_style_args=viewer_style_args
    )

    return viewer
